// MeTA~Fract :: 2002-2004 Tito Dal Canton, tito@dalcanton.it
#include <stdlib.h>
#include <stdio.h>
#include <SDL.h>
#include "mandelbrot.h"
#include "graph-io.h"

#define ALGORITMO_BOUNDARY_TRACE	1
#define PUT_PIXEL(x, y, c)	*(((unsigned char *)curr_surface->pixels) + (y) * curr_surface->pitch + (x)) = c
#define GET_PIXEL(x, y)		*(((unsigned char *)curr_surface->pixels) + (y) * curr_surface->pitch + (x))

enum {
	kPixelNotDone = 0,  // il pixel non e' stato calcolato
	kPixelJustDone,     // il pixel non era calcolato, lo abbiamo calcolato adesso
	kPixelDone          // il pixel era gia' calcolato
};

static void GeneratePixel(int xp, int yp, int update);
static unsigned char PixelStatus(int x, int y);
static int BTraceColorZone(void);

SDL_Surface			*curr_surface = NULL;
extern int			width, height;
extern int			whichPalette;
static int			gFractalWidth = 0, gFractalHeight = 0;
extern double		gSupX, gInfX, gSupY, gInfY;
static double		incremx, incremy;
extern int			gMaxIterations;
int 				gPixelX = 0, gPixelY = 0;
unsigned char		*gPixelStatusBuf = NULL;	// 'stato' di ogni pixel

// riporta il piano complesso ai valori iniziali, in modo da
// avere una visione completa dell'intero insieme
void FullMandelbrotSet(int w, int h)
{
	double	aspect;
	
	gFractalWidth = w;
	gFractalHeight = h;
	aspect = (double)gFractalHeight / (double)gFractalWidth;
    if (gFractalWidth > gFractalHeight) {
		gInfY = -2; gSupY = 2;
		gInfX = gInfY / aspect;
		gSupX = gSupY / aspect;
	} else {
		gInfX = -2; gSupX = 2;
		gInfY = aspect * gInfX;
		gSupY = aspect * gSupX;
	}
}

// distruggi e rialloca il buffer che mantiene lo stato di ogni pixel
int ReallocPixelStatusBuf(void)
{
	if (gPixelStatusBuf != NULL)
		free((char *)gPixelStatusBuf);
	gPixelStatusBuf = (unsigned char *)malloc(gFractalWidth * gFractalHeight);
	if (gPixelStatusBuf == NULL)
		return -1;
    return 0;
}

// disegna insieme di Mandelbrot e aree equipotenziali con colori diversi,
// tramite l'elegantizzimo algoritmo che traccia i contorni delle zone dello stesso
// colore, e poi le riempie senza calcolare il colore di ogni loro pixel.
void RenderMandelbrot(SDL_Surface *surf, int algo)
{
	int				repeat = 0;
	unsigned long	cellwidth = 128;
	unsigned long	prima, dopo;
	
	printf("Posizione nel piano complesso: %f,%f %f,%f\n", gInfX, gInfY, gSupX, gSupY);
	// aggiorna i parametri
	curr_surface = surf;
	gFractalWidth = surf->w;
	gFractalHeight = surf->h;
	gPixelX = gPixelY = 0;
	incremx = (gSupX - gInfX) / gFractalWidth;
	incremy = (gSupY - gInfY) / gFractalHeight;
	ReallocPixelStatusBuf();
	
	SDL_LockSurface(surf);
	// cancella tutto lo stato dei pixel e la finestra
	memset(gPixelStatusBuf, kPixelNotDone, gFractalWidth * gFractalHeight);
	memset(surf->pixels, 1, gFractalWidth * gFractalHeight);
	SDL_UpdateRect(surf, 0, 0, gFractalWidth, gFractalHeight);
	
	// renderizza il frattale
	prima = SDL_GetTicks();
	if (algo == BOUNDARY_TRACE) {
		// superveloce per grandi zone di colore uniforme
		do {
			SDL_Event	event;
			
			repeat = BTraceColorZone();
			event.type = SDL_NOEVENT;
			SDL_PollEvent(&event);
			if (event.type == SDL_KEYDOWN) {
				switch (event.key.keysym.sym) {
					case SDLK_ESCAPE:
						repeat = 0;
						break;
					case SDLK_p:
						whichPalette++;
						if (whichPalette >= PALETTE_COUNT)
							whichPalette = 0;
						ChangePalette(whichPalette, curr_surface);
						break;
					default:
						break;
				}
			}
		} while (repeat);
	} else if (algo == APPROXIMATION) {
		// approssimazioni successive - pi lento e accurato
		// (adatto alle esplorazioni di zome molto "caotiche")
		int	stop_rendering = 0;
		
		while (cellwidth >= 1 && !stop_rendering) {
			int	x, y;

			for (y = 0; y < gFractalHeight; y += cellwidth) {
				int			first = 0;
				SDL_Event	event;

				if ((y & cellwidth) == 0)
					first = cellwidth;
				for (x = first; x < gFractalWidth; x += cellwidth) {
					unsigned char	status;
					
					if (first != 0 && (x & cellwidth) != first)
						continue;
					GeneratePixel(x, y, 0);
					if (cellwidth > 1) {
						int			color = GET_PIXEL(x, y);
						SDL_Rect    r = {x, y, cellwidth, cellwidth};
					
						SDL_FillRect(surf, &r, color);
					}
				}
				SDL_UpdateRect(surf, 0, y, gFractalWidth, cellwidth);
				// gestisci gli eventi
				event.type = SDL_NOEVENT;
				SDL_PollEvent(&event);
				if (event.type == SDL_KEYDOWN) {
					switch (event.key.keysym.sym) {
						case SDLK_ESCAPE:
							stop_rendering = 1;
							y = gFractalHeight;
							break;
						case SDLK_p:
							whichPalette++;
							if (whichPalette >= PALETTE_COUNT)
								whichPalette = 0;
							ChangePalette(whichPalette, curr_surface);
							break;
						default:
							break;
					}
				}
			}
			cellwidth >>= 1;
		}
	} else if (algo == PIXEL_BY_PIXEL) {
		// algoritmo stupidissimo pixel-per-pixel,
		// usato nei rendering su file per avere la certezza
		// di un'immagine esatta (uber-paranoia :) e per
		// mostrare la barra di avanzamento sullo schermo
		int			x, y;
		float		precalc = (main_surface->w - 12.0) / gFractalHeight;
		SDL_Rect	bar_rect = {5, 72, main_surface->w - 10, 20};

		SDL_FillRect(main_surface, &bar_rect, 0);
		SDL_UpdateRects(main_surface, 1, &bar_rect);
		bar_rect.w = 1;
		bar_rect.y = 73;
		bar_rect.h = 18;
		for (y = 0; y < gFractalHeight; y++) {
			for (x = 0; x < gFractalWidth; x++)
				GeneratePixel(x, y, 0);
			// aggiorna la barra di avanzamento ;-)
			bar_rect.x = (y * precalc) + 6;
			SDL_FillRect(main_surface, &bar_rect, 1);
			SDL_UpdateRects(main_surface, 1, &bar_rect);
		}
		SDL_UpdateRect(surf, 0, 0, 0, 0);
	}
	
	dopo = SDL_GetTicks();
	printf("**** TEMPO DI RENDERING: %.2f SECONDI ****\n", (dopo - prima) / 1000.0);
	SDL_UnlockSurface(surf);
}

// calcola e disegna un singolo pixel del frattale
static void GeneratePixel(int xp, int yp, int update)
{
	int		c = 0, color;
	double	r = 0.0, r2 = 0.0, i = 0.0, i2 = 0.0, x, y;
	
	x = gInfX + incremx * xp;
	y = gInfY + incremy * yp;

	// calcola il colore del pixel tramite iterazioni della
	// formula di Mandelbrot Zn = Z(n-1)^2 + Z0
	while (((r2 + i2) < 4.0) && (c <= gMaxIterations)) {
		i = 2 * i * r + y;
		r = r2 - i2 + x;
		r2 = r * r;
		i2 = i * i;
		c++;
	}

	if (c < (gMaxIterations - 1))
		color = (c % (kNumOfColors - 2) + 2);
	else
	 	color = 0;
	PUT_PIXEL(xp, yp, color);
	// SDL_UpdateRect puo' essere MOLTO lenta,
	// quindi per certi algoritmi ne facciamo a meno
	if (update)
		SDL_UpdateRect(curr_surface, xp, yp, 1, 1);
}

// ritorna lo stato del pixel specificato, in caso calcolalo
// kPixelDone		-> pixel gia' calcolato e disegnato
// kPixelNotDone	-> pixel non calcolato
// kPixelJustDone	-> pixel calcolato ora
static unsigned char PixelStatus(int x, int y)
{
	if ((x < 0) || (x >= gFractalWidth) || (y < 0) || (y >= gFractalHeight)) {
		return kPixelDone;
	} else {
		unsigned char	*statusptr = (gPixelStatusBuf + x + (y * gFractalWidth));
		
		if (*statusptr == kPixelNotDone) {	// Se il pixel non e' stato fatto...
			GeneratePixel(x, y, 1);			// ...calcolalo...
			*statusptr = kPixelJustDone;	// e segnalo come appena fatto
		}
		return *statusptr;
	}
}

// giungi fino al primo pixel non calcolato, e traccia tutto il bordo dell'area
// con lo stesso colore, quindi riempila senza calcolare tutti i suoi pixel
static int BTraceColorZone(void)
{
	unsigned char	pixStatus;
	int				minx = gFractalWidth, miny = gFractalHeight, maxx = 0, maxy = 0;
	
	// trova il primo pixel non ancora calcolato
	do {
		pixStatus = PixelStatus(gPixelX, gPixelY);
		if (pixStatus == kPixelDone) {
			gPixelX++;
			if (gPixelX == gFractalWidth) {
				gPixelX = 0;
				gPixelY++;
			}
		}
	} while ((pixStatus == kPixelDone) && (gPixelY < gFractalHeight));
	// da questo pixel comincia a tracciare tutto il bordo della zona di colore uniforme
	if (gPixelY != gFractalHeight) {
		unsigned int	color, orgcolor;
		unsigned char	direction = 1;
		unsigned long	justone = 0;
		int				currPixelX = gPixelX, currPixelY = gPixelY;
		
		orgcolor = GET_PIXEL(gPixelX, gPixelY);	// leggi il colore che avra' l'area
		minx = gFractalWidth;
		miny = gFractalHeight;
		maxx = maxy = -1;
		do {
			unsigned char	orgSubDirection, subDirection;
			int				subx, suby;
			
			orgSubDirection = (direction - 1) & 3;
			subDirection = orgSubDirection;
			// questo ciclo calcola le coordinate del prossimo pixel del bordo
			do {
				switch (subDirection) {
					case 0:	subx = 0;	suby = -1;	break;	// pixel sopra
					case 1:	subx = 1;	suby = 0;	break;	// pixel a destra
					case 2:	subx = 0;	suby = 1;	break;	// pixel sotto
					case 3:	subx = -1;	suby = 0;	break;	// pixel a sinistra
				}
				// leggi lo stato di questo pixel
				pixStatus = PixelStatus(currPixelX + subx, currPixelY + suby);
				// se e' appena stato calcolato, leggine il colore
				if (pixStatus == kPixelJustDone)	
					color = GET_PIXEL(currPixelX + subx, currPixelY + suby);
				else
					color = -1;
				if (color != orgcolor)		// se e' di un colore divero dal bordo allora il bordo non prosegue
					pixStatus = kPixelDone;	// in questa direzione, quindi marca il pixel come calcolato e basta
				subDirection++;
				subDirection &= 3;
				if (subDirection == orgSubDirection)
					pixStatus = 3;			// esci se hai esaminato tutti i quattro pixel intorno
			} while (pixStatus == kPixelDone); 
			// adesso abbiamo trovato il pixel successivo del bordo
			if ((pixStatus == kPixelJustDone) || (pixStatus == 3 && justone)) {
				direction = (subDirection - 1) & 3;
				currPixelX += subx;		// partendo da questo nuovo pixel, calcola il prossimo
				currPixelY += suby;
				justone++;
			}
			if (currPixelX < minx)		// aggiorna le dimensioni
				minx = currPixelX;		// del rettangolo che
			if (currPixelY < miny)		// comprende l'area
				miny = currPixelY;		// che stiamo delimitando,
			if (currPixelX > maxx)		// ci serviranno dopo
				maxx = currPixelX;		// per riempirla di colore
			if (currPixelY > maxy)
				maxy = currPixelY;
		} while (currPixelX != gPixelX || currPixelY != gPixelY);
	
		// ok, l'area che cercavamo ora e' chiusa, 'allagala' con il suo colore
		if (justone) {
			int		fillx, filly;
			
			for (filly = miny; filly <= maxy; filly++) {
				unsigned char	*statusRowPtr = gPixelStatusBuf + filly * gFractalWidth;
				int				fill = 0;	// questa indica se stiamo colorando o no
				
				for (fillx = minx; fillx <= maxx; fillx++) {
					pixStatus = *(statusRowPtr + fillx);
					if (pixStatus == kPixelJustDone)
						*(statusRowPtr + fillx) = kPixelDone;
					if (fill) {
						if (pixStatus == kPixelNotDone) {
							PUT_PIXEL(fillx, filly, orgcolor);
							*(statusRowPtr + fillx) = kPixelDone;
						} else {
							color = GET_PIXEL(fillx, filly);
							if (color != orgcolor)
								fill = 0;
							*(statusRowPtr + fillx) = kPixelDone;
						}	
					} else {
						// se incontri un punto del bordo, inizia a riempire la riga
						color = GET_PIXEL(fillx, filly);
						if (pixStatus == kPixelJustDone && color == orgcolor)
							fill = 1;
					}
				}
			}
			if (curr_surface == main_surface)
				SDL_UpdateRect(curr_surface, minx, miny, maxx-minx+1, maxy-miny+1);
		} else {
			*(gPixelStatusBuf + gPixelX + (gPixelY * gFractalWidth)) = kPixelDone;
			if (curr_surface == main_surface)
				SDL_UpdateRect(curr_surface, gPixelX, gPixelY, 1, 1);
		}
		return 1;
	} else {
		return 0;
	}
}

// controlla se due pixel adiacenti presentano la stessa posizione
// nel piano complesso. Se si, la precisione del disegno va a farsi fottere.
// NOTA: va chiamata dopo Fractal_BoundaryTrace
int VisiblePrecisionLoss(void)
{
	if ((gInfX + 7 * incremx) == (gInfX + 8 * incremx))
		return 1;
	return 0;
}

